var mouseFunctions = mouseFunctions || {};

mouseFunctions.checkForClearPassage = function( cells ){

	var aCell;

	// set all currentIteration arrays to their travelTo state.
	for( var i=0, ii=cells.list.length; i<ii; i++ ){
		aCell = cells.list[ i ];
		for( var j=0; j<8; j++ ){
			aCell.currentIteration[ j ] = aCell.travelTo[ j ];
		}
	}

	// first we'll pretend that the wall has been added.
	// then we'll test to see if the path is clear
	updateCell( cells , false , "currentIteration" );

	// path will be an array that we add to.  At each iteration of our Flood Fill pathfinding algorithm
	// the path will copy its self for each subsequent iteration and add a unique bit to each new construct.  
	// Each Path item will contain an array of all spaces it has traveled to and begin with the overall distance.
	// For now the path will start at the last square.
	// the first entry in the path array will be an object indicating whether or not an end goal has been found.
	// and what the indexed position is for the shortest path and the length of the shortest path.
	var path = [ { goal: false , index: -1, length: 10000 } , [ 0 , ( cells.list.length - 1 ) ] ];

	iterateThroughAllDirections( cells , path );

	return path[0].goal;
}

mouseFunctions.createWall = function( index ){

	// the wall does not exist and there is room to build.  
	// we'll create one at this location.
	walls.lookUpArray[ cells.previousSelection[0] ] = true;

	// assign two temporary placeholder variables
	var v1 = grid.geometry.vertices[ cells.previousSelection[0] ];
	var v2 = grid.geometry.vertices[ cells.previousSelection[1] ];

	walls.geometry.vertices[ index ].set( v1.x , v1.y , v1.z );
	walls.geometry.vertices[ index+1 ].set( v1.x , v1.y+0.02 , v1.z );

	walls.geometry.vertices[ index+2 ].set( v2.x , v2.y , v2.z );
	walls.geometry.vertices[ index+3 ].set( v2.x , v2.y+0.02 , v2.z );

	// Now we update the travelTo array to restrict movement.
	updateCell( cells , false , "travelTo" );

	// now that we've set the location of the vertices we'll need to update the geometry.
	// the faces will automatically stretch into place as each set of 2 faces is connected to
	// 4 vertices.
	walls.geometry.verticesNeedUpdate = true;
	walls.geometry.facesNeedUpdate = true;
	walls.geometry.normalsNeedUpdate = true;
	walls.geometry.computeFaceNormals();
	walls.geometry.computeBoundingSphere();
	walls.updateMatrixWorld();
}

mouseFunctions.createTower = function( underList , extraInfo ){

	// we're going to create a line and capture it's center point for placing the tower.
	var line3 = new THREE.Line3( 
		grid.geometry.vertices[ cells.previousSelection[ 0 ] ].clone(),
		grid.geometry.vertices[ cells.previousSelection[ 9 ] ].clone() 
	);

	var centerVector = line3.at( 0.5 );

	var basicMaterial = new THREE.MeshLambertMaterial( { vertexColors: THREE.FaceColors } );
	var tower = { 
		base: new THREE.Mesh( models.import( gameParameters.tower.modelBase , 10 ) , basicMaterial ),
		turret : new THREE.Mesh( models.import( gameParameters.tower.modelTurret , 10 ) , basicMaterial ),
	};
	
	// scale and position tower's base and turret.
	tower.base.scale.set( 0.01 , 0.01 , 0.01 );
	tower.base.position.set( centerVector.x , centerVector.y , centerVector.z );
	tower.base.rotateY( Math.random() * Math.PI );
	tower.base.visible = true;

	tower.turret.scale.set( 0.01 , 0.01 , 0.01 );
	tower.turret.position.set( centerVector.x , centerVector.y , centerVector.z );
	tower.turret.rotateY( Math.random() * Math.PI );
	tower.turret.visible = true;

	towers.push( tower );
	scene.add( towers[ towers.length-1 ].base );
	scene.add( towers[ towers.length-1 ].turret );

	// next we'll change the colors of the grid lines under the placed tower
	// so that it will look like they've disappeared.
	var baseColor; 
	for( var i=0, ii=underList.length; i<ii; i++ ){
		baseColor = grid.geometry.colors[ underList[ i ] ].base;
		baseColor.r = 0; 
		baseColor.g = 0; 
		baseColor.b = 0;
		grid.geometry.colors[ underList[ i ] ].setRGB( baseColor.r , baseColor.g , baseColor.b );
	}

	// we're going to create a wall around each tower.
	for( var i=0, ii=cells.previousSelection.length; i<ii; i+=2 ){

		// regardless of whether the wall exists here, we're going to create a wall around
		// the tower.  so we'll need to create 8 new walls.
		towerWalls.lookUpArray[ cells.previousSelection[i] ] = true;

		// assign two temporary placeholder variables
		var v1 = grid.geometry.vertices[ cells.previousSelection[i] ];
		var v2 = grid.geometry.vertices[ cells.previousSelection[i+1] ];

		towerWalls.geometry.vertices[ towerWalls.bufferIndex.vertices ].set( v1.x , v1.y , v1.z );
		towerWalls.geometry.vertices[ towerWalls.bufferIndex.vertices+1 ].set( v1.x , v1.y+0.02 , v1.z );

		towerWalls.geometry.vertices[ towerWalls.bufferIndex.vertices+2 ].set( v2.x , v2.y , v2.z );
		towerWalls.geometry.vertices[ towerWalls.bufferIndex.vertices+3 ].set( v2.x , v2.y+0.02 , v2.z );

		// add an entry into buffer Link
		towerWalls.bufferLink[ cells.previousSelection[0] ] = towerWalls.bufferIndex.vertices;

		// lastly, update the buffer index
		towerWalls.bufferIndex.vertices +=4;
	}

	// now that we've set the location of the vertices we'll need to update the geometry.
	// the faces will automatically stretch into place as each set of 2 faces is connected to
	// 4 vertices.
	towerWalls.geometry.verticesNeedUpdate = true;
	towerWalls.geometry.facesNeedUpdate = true;
	towerWalls.geometry.normalsNeedUpdate = true;
	towerWalls.geometry.computeFaceNormals();
	towerWalls.geometry.computeBoundingSphere();
	towerWalls.updateMatrixWorld();


	// finally we'll update the surrounding cell travelTo Information.
	// first we'll add underList to previousSelection.  Then when we send it off to 
	// updateCell() it'll update all accordingly.
	for( var i=0, ii=underList.length; i<ii; i++ ){
		cells.previousSelection.push( underList[ i ] );
	}
	// we'll add the faceIndex and side info to cells.info[];
	for( var i=0, ii=extraInfo.length; i<ii; i++ ){
		cells.info.push( extraInfo[ i ] );
	}

	updateCell( cells , false , "travelTo" );

	console.log( towers );
}

mouseFunctions.updateCell = function( cells , bool , array ){

	updateCell( cells , bool , array );	
}

function updateCell( cells , bool , array ){

	var aCell, bCell, side;
	
	// the reciprical side, 0<->2 , 1<->3.
	var sideReciprical = [ 2 , 3 , 0 , 1 ];

	// -->> PROBABLY NOT NEEDED
	// we will change the passage directly and diagonally for each side
	// we'll record the diagonal index position in the surroundingCells array
	// side:0 has surroundingCells indexes 1,0 and 7.  etc...
	// REFER TO generateGameContent.createCells() for diagram.
	//var indexLookUp = [ { a:7 , b:1 } , { a:1 , b:3 } , { a:3 , b:5 } , { a:5 , b:7 } ];
	// -----<<

	// we will first set the direct passage from one cell to another and then vis-versa with the connect cells
	// passage back.
	// then we'll check for diagonal travel in the next for loop.
	
	for( var i=0, ii=cells.info.length; i<ii; i++ ){
		
		side = cells.info[ i ].side;

		aCell = cells.list[ cells.info[ i ].faceIndex ];
		aCell.currentIteration[ side * 2 ] = false;
		// -->> PROBABLY NOT NEEDED
		//aCell.currentIteration[ indexLookUp[ side ].a ] = false;
		//aCell.currentIteration[ indexLookUp[ side ].b ] = false;
		// -----<< 

		bCell = cells.list[ aCell.surroundingCells[ sideReciprical[ side ] * 2 ] ];
		bCell.currentIteration[ sideReciprical[ side ] * 2 ] = false;
		// -->> PROBABLY NOT NEEDED
		//bCell.currentIteration[ sideReciprical[ indexLookUp[ side ].a ] ] = false;
		//bCell.currentIteration[ sideReciprical[ indexLookUp[ side ].b ] ] = false;
		// -----<<
	}	

	// we need to update any cell that has this wall apart of it's surroundings
	// since we allow the AI to travel diagonally we have to check more conditions 
	// than if we only allowed left/right up/down movement.
	for( var i=0, ii=cells.info.length; i<ii; i++ ){

		switch( cells.info[ i ].side ){
			case 0:

				checkForDiagonalPassage( cells , i , array , 0 , 4 , 2 , 6 , 1 , 3 , 5 , 7 , bool );
				break;

			case 1:

				checkForDiagonalPassage( cells , i , array  , 2 , 6 , 0 , 4 , 1 , 7 , 5 , 3 , bool );
				break;

			case 2:

				checkForDiagonalPassage( cells , i , array  , 4 , 0 , 2 , 6 , 3 , 1 , 7 , 5 , bool );
				break;

			case 3:

				checkForDiagonalPassage( cells , i , array  , 6 , 2 , 0 , 4 , 7 , 1 , 3 , 5 , bool );
				break;
		}
	}
}

function checkForDiagonalPassage( cells , i , array , a , b , c , d , w , x , y , z , bool ){

	var aCell = cells.list[ cells.info[ i ].faceIndex ], bCell, cCell, dCell, eCell, fCell;

	aCell[ array ][ a ] = bool;

	bCell = cells.list[ aCell.surroundingCells[ a ] ];
	bCell[ array ][ b ] = bool;

	cCell = cells.list[ aCell.surroundingCells[ w ] ];
	dCell = cells.list[ aCell.surroundingCells[ z ] ];
	eCell = cells.list[ aCell.surroundingCells[ c ] ];
	fCell = cells.list[ aCell.surroundingCells[ d ] ];

	// check corners and look for a previous wall.  If one is found update all four cells for each
	// corner with false, otherswise, set the travelTo state to true.
	// 
	// 		| b | c |   |
	//		| a | x |   | 
	// 		|   |   |   |
	//
	// 		!( x -> a ) || !( a -> b ) || !( b -> c ){
	//			x->b && b->x && a->c && c->a = false	
	//		} else { 
	//			x->b && b->x && a->c && c->a = true	
	//		} 	

	if( !aCell[ array ][ a ] || !bCell[ array ][ c ] || !cCell[ array ][ b ] ){
		aCell[ array ][ w ] = false;
		bCell[ array ][ x ] = false;
		cCell[ array ][ y ] = false;
		eCell[ array ][ z ] = false;
	} else {
		aCell[ array ][ w ] = true;
		bCell[ array ][ x ] = true;
		cCell[ array ][ y ] = true;
		eCell[ array ][ z ] = true;

	}
	if( !aCell[ array ][ a ] || !bCell[ array ][ d ] || !dCell[ array ][ b ] ){
		aCell[ array ][ z ] = false;
		bCell[ array ][ y ] = false;
		dCell[ array ][ x ] = false;
		fCell[ array ][ w ] = false;
	} else {
		aCell[ array ][ z ] = true;
		bCell[ array ][ y ] = true;
		dCell[ array ][ x ] = true;
		fCell[ array ][ w ] = true;
	}
}

function iterateThroughAllDirections( cells , path ){

	var aCell, bCell, newPath=[];

	// the reciprical cell position of connected cell.
	var cellReciprical = [ 4 , 5 , 6 , 7 , 0 , 1 , 2 , 3 ];

	// cycle through all paths.
	for( var a=1, aa=path.length; a<aa; a++ ){

		aCell = cells.list[ path[ a ][ path[ a ].length - 1 ] ];

		//console.log( aCell );
		// we'll copy over the length of this path segment into the new array newPath.
		newPath = [ path[ a ][ 0 ] ];

		// we're going to store the length of the path in the zero index of newPath
		// if we travel from left/right or up/down we'll increment by 1
		// if diagonally, we'll increment by 1.415

		// check for passage to each of the eight connected cells
		for( var i=0; i<8; i++ ){
			
			bCell = cells.list[ aCell.surroundingCells[ i ] ];

			// if passage is allowed, copy the indexed path array and continue.
			if( aCell.currentIteration[ i ] ){

				newPath = [];

				for( var j=0, jj=path[ a ].length; j<jj; j++ ){
					newPath[ j ] = path[ a ][ j ];
				}

				// we'll add to the length of the current iteration/newPath
				if( i % 2 == 0 ){
					newPath[ 0 ] += 1;
				} else {
					newPath[ 0 ] += 1.415;
				}

				// adjust currentIteration for both to and from cells to false;
				aCell.currentIteration[ i ] = false;
				bCell.currentIteration[ cellReciprical[ i ] ] = false;

				//check to see if this cell is a goal cell
				if( bCell.goal ){

					// we've found a path to the exit.
					path[ 0 ].goal = true;

					// is this newly found path shorter than any previous ones?
					if( newPath[ 0 ] < path[ 0 ].length ){
						path[ 0 ].index = a;
						path[ 0 ].length = newPath[ 0 ];
					}

					// we'll cap this 'path' with the goals location
					path[ a ].push( aCell.surroundingCells[ i ] );
					i=8;
					
				} else {

					// we'll increment another time since we haven't reached the goal.
					newPath.push( aCell.surroundingCells[ i ] );
					path.push( newPath );

					// cause another cycle of the funciton.
					aa++;
				}
			} 
		}
	}
}